module LibASICamera

# Library to use the ASI SDK in a Julia Language application

#
# Adapted from LibASICamera.jl 
#   https://juliapackages.com/p/libasicamera
#   https://juliapackages.com/u/alftetzlaff
# 
# Depends on the ASI SDK libraries installed in ../lib
#   https://www.zwoastro.com/software/

# Version 1.0: 2025-03-08
#   Collected LibASICamera code into one  file
#   Added getters and setters
#   Consistent use of microseconds for exposure time
#   Capture_still looks for success to read the camera


import Libdl
using CEnum

const libASICamera2 = joinpath(@__DIR__, "../lib/libASICamera2")
const ASICAMERA_ID_MAX = 128

# Type and error enumerations

@cenum ASI_CONTROL_TYPE::UInt32 begin
    ASI_GAIN = 0
    ASI_EXPOSURE = 1  # Exposure time in microsec
    ASI_GAMMA = 2     # 1-100, default: 50
    ASI_WB_R = 3      # white balance, red component
    ASI_WB_B = 4      # white balance, blue
    ASI_OFFSET = 5    # pixel value offset / bias
    ASI_BANDWIDTHOVERLOAD = 6  # data transfer rate percentage
    ASI_OVERCLOCK = 7
    ASI_TEMPERATURE = 8  # 10 times the actual temperature
    ASI_FLIP = 9
    ASI_AUTO_MAX_GAIN = 10  # Maximum gain when auto adjust
    ASI_AUTO_MAX_EXP = 11   # Maximum exposure time when auto adjust, millisec
    ASI_AUTO_TARGET_BRIGHTNESS = 12  # Target brightness when auto adjust
    ASI_HARDWARE_BIN = 13
    ASI_HIGH_SPEED_MODE = 14
    ASI_COOLER_POWER_PERC = 15
    ASI_TARGET_TEMP = 16  # °C, don't multiply by 10
    ASI_COOLER_ON = 17
    ASI_MONO_BIN = 18  # lead to a smaller grid at software bin mode for color camera?!
    ASI_FAN_ON = 19
    ASI_PATTERN_ADJUST = 20  # currently only supported by 1600 mono camera
    ASI_ANTI_DEW_HEATER = 21
end

const ASI_BRIGHTNESS = ASI_OFFSET
const ASI_AUTO_MAX_BRIGHTNESS = ASI_AUTO_TARGET_BRIGHTNESS

@cenum ASI_BAYER_PATTERN::UInt32 begin
    ASI_BAYER_RG = 0
    ASI_BAYER_BG = 1
    ASI_BAYER_GR = 2
    ASI_BAYER_GB = 3
end

@cenum ASI_IMG_TYPE::Int32 begin
    ASI_IMG_RAW8 = 0
    ASI_IMG_RGB24 = 1
    ASI_IMG_RAW16 = 2
    ASI_IMG_Y8 = 3
    ASI_IMG_END = -1
end

@cenum ASI_GUIDE_DIRECTION::UInt32 begin
    ASI_GUIDE_NORTH = 0
    ASI_GUIDE_SOUTH = 1
    ASI_GUIDE_EAST = 2
    ASI_GUIDE_WEST = 3
end

@cenum ASI_FLIP_STATUS::UInt32 begin
    ASI_FLIP_NONE = 0
    ASI_FLIP_HORIZ = 1
    ASI_FLIP_VERT = 2
    ASI_FLIP_BOTH = 3
end

@cenum ASI_CAMERA_MODE::Int32 begin
    ASI_MODE_NORMAL = 0
    ASI_MODE_TRIG_SOFT_EDGE = 1
    ASI_MODE_TRIG_RISE_EDGE = 2
    ASI_MODE_TRIG_FALL_EDGE = 3
    ASI_MODE_TRIG_SOFT_LEVEL = 4
    ASI_MODE_TRIG_HIGH_LEVEL = 5
    ASI_MODE_TRIG_LOW_LEVEL = 6
    ASI_MODE_END = -1
end

@cenum ASI_TRIG_OUTPUT::Int32 begin
    ASI_TRIG_OUTPUT_PINA = 0
    ASI_TRIG_OUTPUT_PINB = 1
    ASI_TRIG_OUTPUT_NONE = -1
end

const ASI_TRIG_OUTPUT_PIN = ASI_TRIG_OUTPUT

@cenum ASI_ERROR_CODE::UInt32 begin
    ASI_SUCCESS = 0
    ASI_ERROR_INVALID_INDEX = 1
    ASI_ERROR_INVALID_ID = 2
    ASI_ERROR_INVALID_CONTROL_TYPE = 3
    ASI_ERROR_CAMERA_CLOSED = 4
    ASI_ERROR_CAMERA_REMOVED = 5
    ASI_ERROR_INVALID_PATH = 6
    ASI_ERROR_INVALID_FILEFORMAT = 7
    ASI_ERROR_INVALID_SIZE = 8
    ASI_ERROR_INVALID_IMGTYPE = 9
    ASI_ERROR_OUTOF_BOUNDARY = 10
    ASI_ERROR_TIMEOUT = 11
    ASI_ERROR_INVALID_SEQUENCE = 12
    ASI_ERROR_BUFFER_TOO_SMALL = 13
    ASI_ERROR_VIDEO_MODE_ACTIVE = 14
    ASI_ERROR_EXPOSURE_IN_PROGRESS = 15
    ASI_ERROR_GENERAL_ERROR = 16
    ASI_ERROR_INVALID_MODE = 17
    ASI_ERROR_END = 18
end


struct ASIError <: Exception
    code::ASI_ERROR_CODE
end

Base.showerror(io::IO, err::ASIError) = print(io, err.code)


struct ASIWrapperError <: Exception
    message::String
end

Base.showerror(io::IO, err::ASIWrapperError) = print(io, err.message)

@cenum ASI_BOOL::UInt32 begin
    ASI_FALSE = 0
    ASI_TRUE = 1
end

ASI_BOOL(b::Bool) = b ? ASI_TRUE : ASI_FALSE

@cenum ASI_EXPOSURE_STATUS::UInt32 begin
    ASI_EXP_IDLE = 0
    ASI_EXP_WORKING = 1
    ASI_EXP_SUCCESS = 2
    ASI_EXP_FAILED = 3
end


# Structs
# kept mutable for now, has to be checked
# added constructors

mutable struct _ASI_CAMERA_INFO
    Name::NTuple{64, Cchar}
    CameraID::Cint
    MaxHeight::Clong
    MaxWidth::Clong
    IsColorCam::ASI_BOOL
    BayerPattern::ASI_BAYER_PATTERN
    SupportedBins::NTuple{16, Cint}
    SupportedVideoFormat::NTuple{8, ASI_IMG_TYPE}
    PixelSize::Cdouble
    MechanicalShutter::ASI_BOOL
    ST4Port::ASI_BOOL
    IsCoolerCam::ASI_BOOL
    IsUSB3Host::ASI_BOOL
    IsUSB3Camera::ASI_BOOL
    ElecPerADU::Cfloat
    BitDepth::Cint
    IsTriggerCam::ASI_BOOL
    Unused::NTuple{16, UInt8}
end

_ASI_CAMERA_INFO() = _ASI_CAMERA_INFO(
    NTuple{64, Cchar}([0 for i in 1:64]),
    0,0,0,
    ASI_FALSE,
    ASI_BAYER_RG,
    NTuple{16, Cint}([0 for i in 1:16]),
    NTuple{8, ASI_IMG_TYPE}([ASI_IMG_END for i in 1:8]),
    0,
    ASI_FALSE,
    ASI_FALSE,
    ASI_FALSE,
    ASI_FALSE,
    ASI_FALSE,
    1.,
    0,
    ASI_FALSE,
    NTuple{16, UInt8}([0 for i in 1:16])
)

# human-readable type
struct ASI_CAMERA_INFO
    Name::String
    CameraID::Cint
    MaxHeight::Clong
    MaxWidth::Clong
    IsColorCam::ASI_BOOL
    BayerPattern::ASI_BAYER_PATTERN
    SupportedBins::Vector
    SupportedVideoFormat::Vector
    PixelSize::Cdouble
    MechanicalShutter::ASI_BOOL
    ST4Port::ASI_BOOL
    IsCoolerCam::ASI_BOOL
    IsUSB3Host::ASI_BOOL
    IsUSB3Camera::ASI_BOOL
    ElecPerADU::Cfloat
    BitDepth::Cint
    IsTriggerCam::ASI_BOOL
end

ASI_CAMERA_INFO(info::_ASI_CAMERA_INFO) = ASI_CAMERA_INFO(
    split(String([Char(c) for c in info.Name]), "\0")[1],
    info.CameraID,
    info.MaxHeight,
    info.MaxWidth,
    info.IsColorCam,
    info.BayerPattern,
    [i for i in info.SupportedBins if i > 0],
    [i for i in info.SupportedVideoFormat if i > ASI_IMG_END],
    info.PixelSize,
    info.MechanicalShutter,
    info.ST4Port,
    info.IsCoolerCam,
    info.IsUSB3Host,
    info.IsUSB3Camera,
    info.ElecPerADU,
    info.BitDepth,
    info.IsTriggerCam
)


mutable struct _ASI_CONTROL_CAPS
    Name::NTuple{64, Cchar}
    Description::NTuple{128, UInt8}
    MaxValue::Clong
    MinValue::Clong
    DefaultValue::Clong
    IsAutoSupported::ASI_BOOL
    IsWritable::ASI_BOOL
    ControlType::ASI_CONTROL_TYPE
    Unused::NTuple{32, UInt8}
end

_ASI_CONTROL_CAPS() = _ASI_CONTROL_CAPS(
    NTuple{64, Cchar}([0 for i in 1:64]),   # Name
    NTuple{128, Cchar}([0 for i in 1:128]), # Description
    0, 0, 0,    # Max, Min, Default
    ASI_FALSE,  # IsAutoSupported
    ASI_FALSE,  # IsWritable
    ASI_GAIN,   # ControlType; gain corresponds to 0
    NTuple{32, UInt8}([0 for i in 1:32])    # Unused
)

# human-readable type
struct ASI_CONTROL_CAPS
    Name::String
    Description::String
    MaxValue::Clong
    MinValue::Clong
    DefaultValue::Clong
    IsAutoSupported::ASI_BOOL
    IsWritable::ASI_BOOL
    ControlType::ASI_CONTROL_TYPE
end

ASI_CONTROL_CAPS(control_caps::_ASI_CONTROL_CAPS) = ASI_CONTROL_CAPS(
    split(String([Char(c) for c in control_caps.Name]), "\0")[1],   # Name
    split(String([Char(c) for c in control_caps.Description]), "\0")[1], # Description
    control_caps.MaxValue, control_caps.MinValue, control_caps.DefaultValue,    # Max, Min, Default
    control_caps.IsAutoSupported,  # IsAutoSupported
    control_caps.IsWritable,       # IsWritable
    control_caps.ControlType       # ControlType; gain corresponds to 0
)

# const ASI_CONTROL_CAPS = _ASI_CONTROL_CAPS

mutable struct _ASI_ID
    id::NTuple{8, Cuchar}
end

_ASI_ID() = _ASI_ID(NTuple{8, Cuchar}(zeros(Cuchar, 10)))

const ASI_ID = _ASI_ID
const ASI_SN = ASI_ID

mutable struct _ASI_SUPPORTED_MODE
    SupportedCameraModes::NTuple{16, ASI_CAMERA_MODE}
end

_ASI_SUPPORTED_MODE() = _ASI_SUPPORTED_MODE(
    NTuple{16, ASI_CAMERA_MODE}(fill(ASI_MODE_END, 16))
)

const ASI_SUPPORTED_MODE = _ASI_SUPPORTED_MODE

function ASIGetNumOfConnectedCameras()
    ccall((:ASIGetNumOfConnectedCameras, libASICamera2), Cint, ())
end

function ASIGetProductIDs(pPIDs)
    ccall((:ASIGetProductIDs, libASICamera2), Cint, (Ref{Cint},), pPIDs)
end

function ASIGetCameraProperty(pASICameraInfo, iCameraIndex)
    ccall((:ASIGetCameraProperty, libASICamera2), ASI_ERROR_CODE, (Ref{_ASI_CAMERA_INFO}, Cint), pASICameraInfo, iCameraIndex)
end

function ASIGetCameraPropertyByID(iCameraID, pASICameraInfo)
    ccall((:ASIGetCameraPropertyByID, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{ASI_CAMERA_INFO}), iCameraID, pASICameraInfo)
end

function ASIOpenCamera(iCameraID)
    ccall((:ASIOpenCamera, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASIInitCamera(iCameraID)
    ccall((:ASIInitCamera, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASICloseCamera(iCameraID)
    ccall((:ASICloseCamera, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

# ASI_ERROR_CODE ASIGetNumOfControls(int iCameraID, int * piNumberOfControls);
function ASIGetNumOfControls(iCameraID, piNumberOfControls::Ref{Cint})
    ccall((:ASIGetNumOfControls, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cint}), iCameraID, piNumberOfControls)
end

function ASIGetControlCaps(iCameraID, iControlIndex, pControlCaps)
    ccall((:ASIGetControlCaps, libASICamera2), ASI_ERROR_CODE, (Cint, Cint, Ref{_ASI_CONTROL_CAPS}), iCameraID, iControlIndex, pControlCaps)
end

function ASIGetControlValue(iCameraID, ControlType, plValue::Ref{Clong}, pbAuto::Ref{ASI_BOOL})
    ccall((:ASIGetControlValue, libASICamera2), ASI_ERROR_CODE, (Cint, Cint, Ref{Clong}, Ref{ASI_BOOL}), iCameraID, ControlType, plValue, pbAuto)
end

function ASISetControlValue(iCameraID, ControlType, lValue, bAuto)
    ccall((:ASISetControlValue, libASICamera2), ASI_ERROR_CODE, (Cint, Cint, Clong, Cint), iCameraID, ControlType, lValue, bAuto)
end

function ASISetROIFormat(iCameraID, iWidth, iHeight, iBin, Img_type::ASI_IMG_TYPE)
    ccall((:ASISetROIFormat, libASICamera2), ASI_ERROR_CODE, (Cint, Cint, Cint, Cint, ASI_IMG_TYPE), iCameraID, iWidth, iHeight, iBin, Img_type)
end

function ASIGetROIFormat(iCameraID, piWidth::Ref{Cint}, piHeight::Ref{Cint}, piBin::Ref{Cint}, pImg_type::Ref{ASI_IMG_TYPE})
    ccall((:ASIGetROIFormat, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cint}, Ref{Cint}, Ref{Cint}, Ref{ASI_IMG_TYPE}), iCameraID, piWidth, piHeight, piBin, pImg_type)
end

function ASISetStartPos(iCameraID, iStartX, iStartY)
    ccall((:ASISetStartPos, libASICamera2), ASI_ERROR_CODE, (Cint, Cint, Cint), iCameraID, iStartX, iStartY)
end

function ASIGetStartPos(iCameraID, piStartX::Ref{Cint}, piStartY::Ref{Cint})
    ccall((:ASIGetStartPos, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cint}, Ref{Cint}), iCameraID, piStartX, piStartY)
end

function ASIGetDroppedFrames(iCameraID, piDropFrames)
    ccall((:ASIGetDroppedFrames, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cint}), iCameraID, piDropFrames)
end

function ASIEnableDarkSubtract(iCameraID, pcBMPPath)
    ccall((:ASIEnableDarkSubtract, libASICamera2), ASI_ERROR_CODE, (Cint, Cstring), iCameraID, pcBMPPath)
end

function ASIDisableDarkSubtract(iCameraID)
    ccall((:ASIDisableDarkSubtract, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASIStartVideoCapture(iCameraID)
    ccall((:ASIStartVideoCapture, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASIStopVideoCapture(iCameraID)
    ccall((:ASIStopVideoCapture, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASIGetVideoData(iCameraID, pBuffer::T, lBuffSize, iWaitms) where T
    if T == Matrix{UInt8}
        ccall((:ASIGetVideoData, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cuchar}, Clong, Cint), iCameraID, pBuffer, lBuffSize, iWaitms)
    elseif T == Matrix{UInt16}  # explicitly handle unsafe conversion from uint16 to uint8
        ccall((:ASIGetVideoData, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cuchar}, Clong, Cint), iCameraID, Base.unsafe_convert(Ref{Cuchar}, pBuffer), lBuffSize, iWaitms)
    end
end

function ASIPulseGuideOn(iCameraID, direction)
    ccall((:ASIPulseGuideOn, libASICamera2), ASI_ERROR_CODE, (Cint, Cint), iCameraID, direction)
end

function ASIPulseGuideOff(iCameraID, direction)
    ccall((:ASIPulseGuideOff, libASICamera2), ASI_ERROR_CODE, (Cint, Cint), iCameraID, direction)
end

function ASIStartExposure(iCameraID, bIsDark)
    ccall((:ASIStartExposure, libASICamera2), ASI_ERROR_CODE, (Cint, Cint), iCameraID, bIsDark)
end

function ASIStopExposure(iCameraID)
    ccall((:ASIStopExposure, libASICamera2), ASI_ERROR_CODE, (Cint,), iCameraID)
end

function ASIGetExpStatus(iCameraID, pExpStatus)
    ccall((:ASIGetExpStatus, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{ASI_EXPOSURE_STATUS}), iCameraID, pExpStatus)
end

function ASIGetDataAfterExp(iCameraID, pBuffer::T, lBuffSize) where T
    if T == Matrix{UInt8}
        ccall((:ASIGetDataAfterExp, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cuchar}, Clong), iCameraID, pBuffer, lBuffSize)
    elseif T == Matrix{UInt16}  # explicitly handle unsafe conversion from uint16 to uint8
        ccall((:ASIGetDataAfterExp, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cuchar}, Clong), iCameraID, Base.unsafe_convert(Ref{Cuchar}, pBuffer), lBuffSize)
    end
end

function ASIGetID(iCameraID, pID)
    ccall((:ASIGetID, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{ASI_ID}), iCameraID, pID)
end

function ASISetID(iCameraID, ID)
    ccall((:ASISetID, libASICamera2), ASI_ERROR_CODE, (Cint, ASI_ID), iCameraID, ID)
end

function ASIGetGainOffset(iCameraID, pOffset_HighestDR, pOffset_UnityGain, pGain_LowestRN, pOffset_LowestRN)
    ccall((:ASIGetGainOffset, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{Cint}, Ref{Cint}, Ref{Cint}, Ref{Cint}), iCameraID, pOffset_HighestDR, pOffset_UnityGain, pGain_LowestRN, pOffset_LowestRN)
end

function ASIGetSDKVersion()
    ccall((:ASIGetSDKVersion, libASICamera2), Cstring, ())
end

function ASIGetCameraSupportMode(iCameraID, pSupportedMode)
    ccall((:ASIGetCameraSupportMode, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{_ASI_SUPPORTED_MODE}), iCameraID, pSupportedMode)
end

function ASIGetCameraMode(iCameraID, mode)
    ccall((:ASIGetCameraMode, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{ASI_CAMERA_MODE}), iCameraID, mode)
end

function ASISetCameraMode(iCameraID, mode)
    ccall((:ASISetCameraMode, libASICamera2), ASI_ERROR_CODE, (Cint, ASI_CAMERA_MODE), iCameraID, mode)
end

function ASISendSoftTrigger(iCameraID, bStart)
    ccall((:ASISendSoftTrigger, libASICamera2), ASI_ERROR_CODE, (Cint, Cint), iCameraID, bStart)
end

function ASIGetSerialNumber(iCameraID, pSN)
    ccall((:ASIGetSerialNumber, libASICamera2), ASI_ERROR_CODE, (Cint, Ref{ASI_SN}), iCameraID, pSN)
end

function ASISetTriggerOutputIOConf(iCameraID, pin, bPinHigh, lDelay, lDuration)
    ccall((:ASISetTriggerOutputIOConf, libASICamera2), ASI_ERROR_CODE, (Cint, ASI_TRIG_OUTPUT_PIN, Cint, Clong, Clong), iCameraID, pin, bPinHigh, lDelay, lDuration)
end

function ASIGetTriggerOutputIOConf(iCameraID, pin, bPinHigh, lDelay, lDuration)
    ccall((:ASIGetTriggerOutputIOConf, libASICamera2), ASI_ERROR_CODE, (Cint, ASI_TRIG_OUTPUT_PIN, Ref{Cint}, Ref{Clong}, Ref{Clong}), iCameraID, pin, bPinHigh, lDelay, lDuration)
end



"""
    ASICamera

The struct which contains information about the camera.
Has fields .info and .control_caps.
"""
struct ASICamera
    info::ASI_CAMERA_INFO
    control_caps::Vector{ASI_CONTROL_CAPS}
end


"""
    allocate_buffer(width::Integer, height::Integer, img_type::ASI_IMG_TYPE)

Allocates an image buffer for the camera to write to.

# Args:
    width: Image width
    height: Image height
    img_type: One of
        -ASI_IMG_RAW8
        -ASI_IMG_Y8
        -ASI_IMG_RAW16
        -ASI_IMG_RAW24

# Returns:
    A zero-initialized array of the appropriate shape.

# Throws:
    ASIWrapperError if an unsupported image type is given.
"""
function allocate_buffer(width::Integer, height::Integer, img_type::ASI_IMG_TYPE)
    if img_type==ASI_IMG_RAW8 || img_type==ASI_IMG_Y8
        return zeros(UInt8, width, height)
    elseif img_type==ASI_IMG_RAW16
        return zeros(UInt16, width, height)
    elseif img_type==ASI_IMG_RAW24
        return zeros(UInt8, width, height, 3)
    else
        throw(ASIWrapperError("Image type $img_type not implemented."))
    end
end

# This allows to use an ellipsis: allocate_buffer(get_roi_format(cam)...)
allocate_buffer(width::Integer, height::Integer, unused, img_type::ASI_IMG_TYPE) =
    allocate_buffer(width, height, img_type)


"""
    get_num_connected_cameras()

This function returns the count of connected cameras and should be called first.
"""
get_num_connected_cameras() = ASIGetNumOfConnectedCameras()


"""
    get_camera_property(id::Integer)

Fetches the camera properties for a given ID.

# Args:
    id: Camera id

# Returns:
    ASI_CAMERA_INFO object

# Throws:
    ASIError in case of failure
"""
function get_camera_property(id::Integer)
    camera_info = _ASI_CAMERA_INFO()
    err = ASIGetCameraProperty(camera_info, id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
    return ASI_CAMERA_INFO(camera_info) # convert to human-readable type
end

get_camera_property(cam::ASICamera) = get_camera_property(cam.info.CameraID)


"""
    open_camera(id::Integer)

Opens the ASI camera connection.

Open the camera before any interaction with the camera,
this will not affect the camera which is capturing.
Then you must call init_camera() to perform any actions.

# Args:
    id: Camera ID

# Throws:
	ASIError
"""
function open_camera(id::Integer)
    err = ASIOpenCamera(id)
    if err != ASI_SUCCESS
        print("[ERROR] Could not connect to camera. Make sure you can access the camera without being root by calling
        \"sudo install asi.rules /lib/udev/rules.d\" from the lib subdir of the SDK
        and then relogging / rebooting.")
        throw(ASIError(err))
    end
end

open_camera(cam::ASICamera) = open_camera(cam.info.CameraID)


"""
    init_camera(id::Integer)

Initialize the camera. Needs to be called before capturing any data.

# Args:
    id: Camera id

# Throws:
    ASIError
"""
function init_camera(id::Integer)
    err = ASIInitCamera(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

init_camera(cam::ASICamera) = init_camera(cam.info.CameraID)


"""
    close_camera(id::Integer)

Closes the ASI camera connection.

# Args:
    id: Camera id

# Throws:
	ASIError
"""
function close_camera(id::Integer)
    err = ASICloseCamera(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

close_camera(cam::ASICamera) = close_camera(cam.info.CameraID)


"""
    get_control_caps(id::Integer)

Returns the control properties available for this camera.

The camera needs to be open.

# Args:
    id: Camera id

# Returns:
    Vector of ASI_CONTROL_CAPS structs.

# Throws:
    ASIError
"""
function get_control_caps(id::Integer)
    num_controls = Ref{Cint}(0)

    err = ASIGetNumOfControls(id, num_controls)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    control_caps = Vector{ASI_CONTROL_CAPS}(undef, num_controls.x)

    for i in 1:num_controls.x
        cap = _ASI_CONTROL_CAPS()
        err = ASIGetControlCaps(id, i-1, cap)
        if err != ASI_SUCCESS
            throw(ASIError(err))
        end
        control_caps[i] = ASI_CONTROL_CAPS(cap) # convert to human-readable type
    end

    return control_caps
end

get_control_caps(cam::ASICamera) = get_control_caps(cam.info.CameraID)


"""
    get_connected_devices()

Returns a list of the connected devices.
"""
function get_connected_devices()
    num_cameras = get_num_connected_cameras()
    if num_cameras == 0
        throw(ASIWrapperError("No cameras found."))
    end

    cameras = Vector{ASICamera}(undef, num_cameras)
    for i in 0:num_cameras-1
        open_camera(i)
        init_camera(i)
        info = get_camera_property(i)
        control_caps = get_control_caps(i)
        cameras[i+1] = ASICamera(info, control_caps)
    end
    return cameras
end


"""
    get_control_value(id::Integer, control_type::ASI_CONTROL_TYPE)

Fetches the current setting of the control value, e.g. exposure or gain.

# Args:
    id: Camera id
    control_type: The control type to fetch, e.g. exposure or gain.

# Returns:
    A tuple (value, is_auto)

# Throws:
    ASIError
"""
function get_control_value(id::Integer, control_type::ASI_CONTROL_TYPE)
    value = Ref{Clong}(0)
    auto  = Ref{ASI_BOOL}(ASI_FALSE)
    err = ASIGetControlValue(id, control_type, value, auto)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
    return (value[], Bool(auto[]))
end

get_control_value(cam::ASICamera, control_type::ASI_CONTROL_TYPE) = get_control_value(cam.info.CameraID, control_type)


"""
    set_control_value(id::Integer, control_type::ASI_CONTROL_TYPE, value, auto::Bool=false)

Sets a control (e.g. exposure) to the given value. Automatically sets the
minimum or maximum if the given value is out of bounds.

# Args:
    id: Camera id
    control_type: The control type to set, e.g. exposure or gain.
    value: The value to which the control is set.
    auto: Whether or not the control should be automatically set.
        Check if this is supported for the given control beforehand.

# Throws:
    ASIError
"""
function set_control_value(id::Integer, control_type::ASI_CONTROL_TYPE, value, auto::Bool=false)
    err = ASISetControlValue(id, control_type, value, auto)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

set_control_value(cam::ASICamera, control_type::ASI_CONTROL_TYPE, value, auto::Bool=false) = set_control_value(cam.info.CameraID, control_type, value, auto)
set_gain(id, gain, auto=false)                  = set_control_value(id, ASI_GAIN, gain, auto)
set_exposure(id, exposure_microseconds, auto=false)  = set_control_value(id, ASI_EXPOSURE, exposure_microseconds, auto)
set_gamma(id, gamma, auto=false)                = set_control_value(id, ASI_GAMMA, gamma, auto)
set_bandwidth(id, bandwidth, auto=false)        = set_control_value(id, ASI_BANDWIDTHOVERLOAD, bandwidth, auto)
set_flip(id, flip)                              = set_control_value(id, ASI_FLIP, flip)
set_autoexp_max_gain(id, val)                   = set_control_value(id, ASI_AUTO_MAX_GAIN, val)
set_autoexp_max_exp(id, exposure_milliseconds)  = set_control_value(id, ASI_AUTO_MAX_EXP, floor(Int, exposure_milliseconds))
set_autoexp_target_brightness(id, brightness)   = set_control_value(id, ASI_AUTO_MAX_BRIGHTNESS, brightness)
set_highspeed_mode(id, active)                  = set_control_value(id, ASI_HIGH_SPEED_MODE, active)

"""
    get_exposure(id)
    
Returns exposure time as an integer in microseconds
    
"""
get_exposure(id)       = get_control_value(id, ASI_EXPOSURE) 


# ...


"""
    get_status(id::Integer)

Returns the status of all camera parameters.
"""
function get_status(id::Integer)
    control_caps = get_control_caps(id)
    ret = []
    for t in control_caps
        push!(ret, (t.Name, get_control_value(id, t.ControlType)[1]))
    end
    return ret
end

get_status(cam::ASICamera) = get_status(cam.info.CameraID)


"""
    set_roi_format(id::Integer, width, height, binning, img_type::ASI_IMG_TYPE)

Sets the region of interest (roi). Do so before capturing.

The width and height are the values *after* binning, i.e. you need to set the
width to 640 and the height to 480 if you want to run at 640x480 @ BIN2.

ASI120's data size must be a multiple of 1024 which means width*height%1024==0.

# Args:
    id: Camera id
    width: ROI width
    height: ROI height
    binning: The binning mode; 2 means to read out 2x2 pixels together. Check
        which binning values are supported in the ASI_CAMERA_INFO struct of the
        camera struct or by calling get_camera_property(id).

# Throws:
    ASIError
"""
function set_roi_format(id::Integer, width, height, binning, img_type::ASI_IMG_TYPE)
    if width%8 != 0
        throw("Width must be a multiple of 8.")
    end
    if height%2 != 0
        throw("Height must be a multiple of 2.")
    end
    if (width*height)%1024 != 0 && get_camera_property(id).Name in ["ZWO ASI 120MM", "ZWO ASI 120MC"]
        throw("Width times height must be a multiple of 2.")
    end
    err = ASISetROIFormat(id, width, height, binning, img_type)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

set_roi_format(cam::ASICamera, width, height, binning, img_type::ASI_IMG_TYPE) = set_roi_format(cam.info.CameraID, width, height, binning, img_type)


"""
    get_roi_format(id::Integer)

Fetches the current region of interest settings.
"""
function get_roi_format(id::Integer)
    width = Ref{Cint}(0)
    height = Ref{Cint}(0)
    binning = Ref{Cint}(false)
    img_type = Ref{ASI_IMG_TYPE}(ASI_IMG_RAW8)
    err = ASIGetROIFormat(id, width, height, binning, img_type)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
    return (width.x, height.x, binning.x, img_type.x)
end

get_roi_format(cam::ASICamera) = get_roi_format(cam.info.CameraID)


"""
    set_roi_start(id::Integer, startx, starty)

Sets the position of the top-left corner of the region of interest.

You can call this while the camera is streaming to move the ROI. By default,
the ROI will be centered. In binned mode, the start values are relative to the
binned sensor size.

# Throws:
    ASIError
"""
function set_roi_start(id::Integer, startx, starty)
    err = ASISetStartPos(id, startx, starty)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

set_roi_start(cam::ASICamera, startx, starty) = set_roi_start(cam.info.CameraID, startx, starty)


"""
    get_roi_start(id::Integer)

Returns the region of interest start position (start_x, start_y).
"""
function get_roi_start(id::Integer)
    startx = Ref{Cint}(0)
    starty = Ref{Cint}(0)
    err = ASIGGetStartPos(id, startx, starty)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
    return (startx[], starty[])
end

get_roi_start(cam::ASICamera) = get_roi_start(cam.info.CameraID)


"""
    get_dropped_frames(id::Integer)

Returns the number of dropped frames.

Frames are dropped when the USB is bandwidth is low or the
harddisk write speed is slow. The count is reset to 0 after capturing stops.
"""
function get_dropped_frames(id::Integer)
    dropped_frames = Ref{Cint}(0)
    err = ASIGetDroppedFrames(id, dropped_frames)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

get_dropped_frames(cam::ASICamera) = get_dropped_frames(cam.info.CameraID)


"""
"""
function enable_dark_subtract(id::Integer, path)
    err = ASIEnableDarkSubtract(id, path)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

enable_dark_subtract(cam::ASICamera) = enable_dark_subtract(cam.info.CameraID)


"""
"""
function disable_dark_subtract(id::Integer)
    err = ASIDisableDarkSubtract(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

disable_dark_subtract(cam::ASICamera) = disable_dark_subtract(cam.info.CameraID)


"""
    start_video(id::Integer)

Start video capture.

# Throws:
    ASIError
"""
function start_video(id::Integer)
    err = ASIStartVideoCapture(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

start_video(cam::ASICamera) = start_video(cam.info.CameraID)


"""
    stop_video(id::Integer)

Stops a running video capture.

# Throws:
    ASIError
"""
function stop_video(id::Integer)
    err = ASIStopVideoCapture(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

stop_video(cam::ASICamera) = stop_video(cam.info.CameraID)


"""
    get_video_data!(id::Integer, buffer, timeout_ms)

Writes a video frame to the given buffer. Make sure the buffer is large
enough to fit the frame. Call this as fast as possible in a loop and check
whether the return value equals ASI_SUCCESS.

# Args:
    id: Camera id
    buffer: A buffer to write the video frame to.
    timeout_ms: Time to wait for a frame. Recommendation: 2 * exposure_microsec + 500 ms <- inconsistent units?!

# Returns:
    An ASI_ERROR_CODE, which should be ASI_SUCCESS.
"""
function get_video_data!(id::Integer, buffer, timeout_ms=500)
    return ASIGetVideoData(id, buffer, sizeof(buffer), Int32(round(timeout_ms)))
end

get_video_data!(cam::ASICamera, buffer, timeout_ms) = get_video_data!(cam.info.CameraID, buffer, timeout_ms)


"""
    pulse_guide_on(id::Integer, direction::ASI_GUIDE_DIRECTION)

Activates the pulse guide in the given direction.

# Args:
    id: Camera id
    direction: Guiding direction; call 'instances(ASI_GUIDE_DIRECTION)'

# Throws:
    ASIError
"""
function pulse_guide_on(id::Integer, direction::ASI_GUIDE_DIRECTION)
    err = ASIPulseGuideOn(id, direction)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

pulse_guide_on(cam::ASICamera) = pulse_guide_on(cam.info.CameraID)


"""
    pulse_guide_off(id::Integer, direction::ASI_GUIDE_DIRECTION)

Deactivates the pulse guide in the given direction.

# Args:
    id: Camera id
    direction: Guiding direction; call 'instances(ASI_GUIDE_DIRECTION)' for options.

# Throws:
    ASIError
"""
function pulse_guide_off(id::Integer, direction::ASI_GUIDE_DIRECTION)
    err = ASIPulseGuideOff(id, direction)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

pulse_guide_off(cam::ASICamera) = pulse_guide_off(cam.info.CameraID)


"""
    start_exposure(id::Integer, is_dark=false)

Starts an exposure. All relevant parameters (exposure time, gain) have to be
set beforehand by calling set_control_value(...) or e.g. set_gain(...).
"""
function start_exposure(id::Integer, is_dark=false)
    err = ASIStartExposure(id, is_dark)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

start_exposure(cam::ASICamera, is_dark=false) = start_exposure(cam.info.CameraID, is_dark)


"""
    stop_exposure(id::Integer)

Stops an ongoing exposure.
"""
function stop_exposure(id::Integer)
    err = ASIStopExposure(id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

stop_exposure(cam::ASICamera) = stop_exposure(cam.info.CameraID)


"""
    get_exp_status(id::Integer)

Returns the status of an ongoing exposure.
See 'instances(ASI_EXP_STATUS)'.

# Throws:
    ASIError
"""
function get_exp_status(id::Integer)
    status = Ref{ASI_EXPOSURE_STATUS}(ASI_EXP_IDLE)
    err = ASIGetExpStatus(id, status)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
    return status[]
end

get_exp_status(cam::ASICamera) = get_exp_status(cam.info.CameraID)


"""
    get_data_after_exp!(id::Integer, buffer)

Fetches the data after a successful exposure and writes it into buffer.
"""
function get_data_after_exp!(id::Integer, buffer)
    err = ASIGetDataAfterExp(id, buffer, sizeof(buffer))
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

get_data_after_exp!(cam::ASICamera, buffer) = get_data_after_exp!(cam.info.CameraID, buffer)


"""
    get_id(id::Integer)

Returns the camera id stored in flash, only available for USB3 cameras.
"""
function get_id(id::Integer)
    camera_id = ASI_ID()
    err = ASIGetID(id, camera_id)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

get_id(cam::ASICamera) = get_id(cam.info.CameraID)


"""
    capture_still(id::Integer)

Captures a still image. 
set gain, exposure etc. beforehand using
Set_control_value(...).

# Returns:
    An array containing the image.

# Throws:
    ASIWrapperError if the image format is not supported by the wrapper, or
    ASIError in other unfortunate cases
"""
function capture_still(id::Integer)
    exposure_microsec, _ = get_control_value(id, ASI_EXPOSURE)
    width, height, binning, img_type = get_roi_format(id)

    buffer = allocate_buffer(width, height, img_type)

    start_exposure(id, false)
    sleep(0.01)
    # Use this to wait for the exposure time for long exposures
    #   Exposures taken this way with capture_still cannot be paused
    
    #   if exposure_microsec > 1000
    #     sleep(exposure_microsec/1000000)
    #   end
    
    while get_exp_status(id) == ASI_EXP_WORKING
        sleep(0.01)
    end

    if get_exp_status(id) == ASI_EXP_SUCCESS
        get_data_after_exp!(id, buffer)
        return buffer[end:-1:1, :]
    else
        throw("Exposure failed")
    end
end

capture_still(cam::ASICamera) = capture_still(cam.info.CameraID)


"""
    get_gain_offset(id::Integer)

Get the presets for offset and gain values at different "sweet spots".

# Args:
    id: Camera id

# Returns:
    A dictionary containing:
        1. The offset at highest dynamic range
        2. The offset at unity gain
        3. The gain with lowest readout noise
        4. The offset with lowest readout noise

# Throws:
    ASIError
"""
function get_gain_offset(id::Integer)
    offset_highest_dr = Ref{Cint}(0)  # Offset at highest dynamic range
    offset_unity_gain = Ref{Cint}(0)  # Offset at unity gain
    gain_lowest_rn    = Ref{Cint}(0)  # Gain at lowest readout noise
    offset_lowest_rn  = Ref{Cint}(0)  # Offset at lowest readout noise

    err = ASIGetGainOffset(id, offset_highest_dr, offset_unity_gain, gain_lowest_rn, offset_lowest_rn)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    return Dict(["offset_highest_dr" => offset_highest_dr[],
                 "offset_unity_gain" => offset_unity_gain[],
                 "gain_lowest_rn" => gain_lowest_rn[],
                 "offset_lowest_rn" => offset_lowest_rn[]])
end

get_gain_offset(cam::ASICamera) = get_gain_offset(cam.info.CameraID)


"""
    get_sdk_version()

Returns the SDK version.
"""
function get_sdk_version()
    return unsafe_string(ASIGetSDKVersion())
end


"""
    get_supported_modes(id::Integer)

Returns the supported camera modes, only need to call when the
IsTriggerCam in the CameraInfo is true.

# Throws:
    ASIError
"""
function get_supported_modes(id::Integer)
    supported_modes = ASI_SUPPORTED_MODE()

    err = ASIGetCameraSupportMode(id, supported_modes)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    return [m for m in supported_modes.SupportedCameraModes if m > ASI_MODE_END]
end

get_supported_modes(cam::ASICamera) = get_supported_modes(cam.info.CameraID)


"""
    get_camera_mode(id::Integer)

Get the current camera mode, only needed to call when the IsTriggerCam
in the CameraInfo is true.
"""
function get_camera_mode(id::Integer)
    mode = Ref{ASI_CAMERA_MODE}(ASI_MODE_END)

    err = ASIGetCameraMode(id, mode)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    return mode[]
end

get_camera_mode(cam::ASICamera) = get_camera_mode(cam.info.CameraID)


"""
    set_camera_mode(id::Integer, mode::ASI_CAMERA_MODE)

Set the camera mode, only needed to call when the IsTriggerCam
in the CameraInfo is true.
"""
function set_camera_mode(id::Integer, mode::ASI_CAMERA_MODE)
    err = ASISetCameraMode(id, mode)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

set_camera_mode(cam::ASICamera) = set_camera_mode(cam.info.CameraID)


"""
    send_soft_trigger(id::Integer, start::ASI_BOOL)

From original docs: Send a softTrigger. For edge trigger, it only need to set
true which means send a rising trigger to start exposure. For level trigger,
it need to set true first means start exposure, and set false means stop
exposure. Only needed to call when the IsTriggerCam in the CameraInfo is true.
"""
function send_soft_trigger(id::Integer, start::ASI_BOOL)
    err = ASISendSoftTrigger(id, start)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

send_soft_trigger(cam::ASICamera) = send_soft_trigger(cam.info.CameraID)


"""
    get_serial_number(id::Integer)

Returns the camera serial number.
"""
function get_serial_number(id::Integer)
    sn = ASI_SN()

    err = ASIGetSerialNumber(id, sn)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    return unsafe_string(sn.id)
end

get_serial_number(cam::ASICamera) = get_serial_number(cam.info.CameraID)


"""
    set_trigger_output_config(id::Integer, pin::ASI_TRIG_OUTPUT_PIN, high::ASI_BOOL, delay, duration)

Configure the output pin (A or B) of Trigger port. If duration <= 0,
this output pin will be closed. Only need to call when the IsTriggerCam
in the CameraInfo is true.

# Args:
    id: Camera id
    pin: Select the pin for output
    high: If true, the selected pin will output a high level as a signal
					when it is effective.
    delay: The delay between the camera receiving a trigger signal and the
            output of the valid level. From 0 microsec - 2,000,000,000 microsec.
    duration: The duration of the valid level output. Same range as delay.

"""
function set_trigger_output_config(id::Integer, pin::ASI_TRIG_OUTPUT_PIN, high::ASI_BOOL, delay_microsec, duration_microsec)
    err = ASISetTriggerOutputIOConf(id, pin, high, delay_microsec, duration_microsec)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end
end

set_trigger_output_config(cam::ASICamera) = set_trigger_output_config(cam.info.CameraID)


"""
    get_trigger_output_config(id::Integer)

Get the output pin configuration, only needed to call when the IsTriggerCam
in the CameraInfo is true.
"""
function get_trigger_output_config(id::Integer)
    pin = ASI_TRIG_OUTPUT_NONE
    high = ASI_FALSE
    delay = Ref{Clong}(0)
    duration = Ref{Clong}(0)

    err = ASIGetTriggerOutputIOConf(id, pin, high, delay, duration)
    if err != ASI_SUCCESS
        throw(ASIError(err))
    end

    return Dict(["pin" => pin,
                 "high" => high,
                 "delay" => delay,
                 "duration" => duration])
end

get_trigger_output_config(cam::ASICamera) = get_trigger_output_config(cam.info.CameraID)



"""
  set_fan_on(id)

"""  
set_fan_on(id)  = set_control_value(id, ASI_FAN_ON, ASI_TRUE)


"""
  set_fan_off(id)

"""
set_fan_off(id)  = set_control_value(id, ASI_FAN_ON, ASI_FALSE)


"""
  set_cooler_on(id)

"""
set_cooler_on(id)  = set_control_value(id, ASI_COOLER_ON, ASI_TRUE)

  
"""
  set_cooler_off(id)

""" 
set_cooler_off(id)  = set_control_value(id, ASI_COOLER_ON, ASI_FALSE)


"""
  set_cooler_target(id, target_temperature)

target_temperature is  >= -40

"""
set_cooler_target(id, temperature) = set_control_value(id, ASI_TARGET_TEMP, floor(Int, temperature))

"""
  get_cooler_target(id)

"""
get_cooler_target(id) = get_control_value(id, ASI_TARGET_TEMP)[1]


"""
  get_cooler_on(id)

"""
get_cooler_on(id)  = get_control_value(id, ASI_COOLER_ON)[1]


"""
  get_fan_on(id)


"""
get_fan_on(id)  = get_control_value(id, ASI_FAN_ON)[1]

  

"""
  get_cooler_status(id)
  
Returns fan_on, cooler_on, target_temperature (integer >= -40),  sensor_temperature

"""
function get_cooler_status(id)

  fan_on_flag = get_control_value(id, ASI_FAN_ON)[1]
  cooler_on_flag = get_control_value(id, ASI_COOLER_ON)[1]
  target_temperature = get_control_value(id, ASI_TARGET_TEMP)[1]
  sensor_temperature = get_control_value(id, ASI_TEMPERATURE)[1] * 0.1
  return fan_on_flag, cooler_on_flag, target_temperature, sensor_temperature
end
  
  
"""
  get_temperature(id)

""" 
get_temperature(id) = get_control_value(id, ASI_TEMPERATURE)[1] * 0.1




export
   ASICamera,
   open_camera,
   close_camera,
   init_camera,
   start_exposure,
   stop_exposure,
   start_video,
   stop_video,
   allocate_buffer,
   enable_dark_subtract,
   disable_dark_subtract,
   pulse_guide_on,
   pulse_guide_off,
   send_soft_trigger,
   capture_still

# export the rest
foreach(names(@__MODULE__, all=true)) do s
   if startswith(string(s), "ASI") || startswith(string(s), "set") ||startswith(string(s), "get")
       @eval export $s
   end
end

end # module

