#!/home/john/local/bin/julia

# Use an ASI ZWO camera for scientific image acquisition
#
#   Provides a GTK user interface to data acquisition
#   Exports FITS images
#

# Version 1.0: 2025-02-19 
#   Interface is actively updated with temperature and exposure data
#   Basic functions present to set exposure, file names, and camera temperature




###
#
# Load the ASI library wrapper with the installed ASI SDK library
#
# The file name must have an absolute directory structure
#
###

asi_image_lib_bin = "/home/john/julia/asi_image/bin/asi_image_lib.jl"
include(asi_image_lib_bin)  


# Load dependencies

using FITSIO               # FITS management
using Dates                # Date and time
using Gtk4, Gtk4.GLib      # GTK interface


# Camera matters

camera_id = "ZWO ASI 6200"
image_size = (9600,4800,16)
temperature_now = 0.0
temperature_set = -10.0
cool_flag = false
filter_name = "rp"
exposure = 60.0
count = 1
count_max = 500
file_base = "image_"
file_folder = "./"
file_name = file_base * string(count)


#
# These library functions may be used to build this camera controller
# For a complete list see the library function code
#

#   get_num_connected_cameras() returns an integer count
#   get_connected_devices() returns a list of connected devices
#   get_camera_property(id::Integer) returns ASI_CAMERA_INFO for this camera
#   get_serial_number(id::Integer) returns camera serial number
#   get_control_caps(id::Integer) returns ASI_CONTROL_CAPS for this camera 
#   get_control_value(id::Integer, control_type::ASI_CONTROL_TYPE) returns
#     a tuple (value, is_auto)
#   set_control_value(id::Integer, control_type::ASI_CONTROL_TYPE, value, 
#     auto::Bool=false)
#   set_gain(id, gain, auto=false) = 
#     set_control_value(id, ASI_GAIN, gain, auto)
#   set_exposure(id, exposure_seconds, auto=false)   =   
#     set_control_value(id,ASI_EXPOSURE, floor(Int, exposure_seconds * 1000), auto) 
#   set_roi_format(id::Integer, width, height, binning, img_type::ASI_IMG_TYPE)
#   get_roi_format(id::Integer)
#   set_roi_start(id::Integer, startx, starty)
#   get_roi_start(id::Integer) returns (startx, starty)
#   get_status(id::Integer)
#   open_camera(id::Integer)
#   close_camera(id::Integer)
#   init_camera(id::Integer)
#   start_exposure(id::Integer, is_dark=false)
#   stop_exposure(id::Integer)
#   get_exp_status(id::Integer) returns ASI_EXP_STATUS
#   pulse_guide_on(id::Integer, direction::ASI_GUIDE_DIRECTION)
#   pulse_guide_off(id::Integer, direction::ASI_GUIDE_DIRECTION)
#   capture_still(id::Integer) returns an array with the image


#
# These functions are fulfilled here using the library and other packages
#

# Camera selection

# Full image size

# Selected centered subframe size 

# Subframe on/off

# Temperature setpoint

# Cooler on/off

# Sensor temperature

# Expose on/off

# Exposure duration 

# Full path to FITS file storage 

# FITS file name for next image

# FITS header FIELD entry text

"""
    write_image_as_fits(out_file, image_data, dateobs_str, exptime_str, filter_str)

Write a FITS file containing a 16-bit image and header entries

"""
function write_image_as_fits(out_file, image_data, dateobs_str, exptime_str, filter_str)
  (width, height) = size(image_data)
  f = FITS(out_file, "w")
  write(f, Float32.(image_data))
  write_key(f[1], "DATE-OBS", dateobs_str)
  write_key(f[1], "EXPTIME", exptime_str)
  write_key(f[1], "FILTER", "clear")
  close(f)
end

"""
    capture_image(exptime)

Capture an image with the ASI camera 

Accepts the exposure time in seconds 
Other camera parameters must be set before starting the exposure
Immediately starts the exposure and waits for its completion
Returns a 16-bit integer raw image


"""
function capture_image(exptime)

  devices = []
  devices = Main.LibASICamera.get_connected_devices()
  cam = devices[1]
  # @show cam.info
  # @show cam.control_caps
  width, height, binning, img_type = Main.LibASICamera.get_roi_format(cam)
  Main.LibASICamera.set_roi_format(cam, 9576, 6388, 1, Main.LibASICamera.ASI_IMG_RAW16)
  Main.LibASICamera.set_exposure(cam, exptime, true)
  image_data = Main.LibASICamera.capture_still(cam)
  Main.LibASICamera.close_camera(cam)
  return image_data
end

"""
    snap(filename)
    
Take a single image and save a fits file by the name in the function call

"""
function snap(filename)
  exptime_str = string(exptime)
  dateobs_str = string(DateTime(now()))
  image_data = capture_image(exptime)
  filter_str = filter
  write_image_as_fits(filename, image_data, dateobs_str, exptime_str, filter_str)
end

# Testing camera and software begins here

devices = []
devices = Main.LibASICamera.get_connected_devices()
cam = devices[1]

@show cam.info
#camera_name   = cam.info
#camera_height = cam.info
#camera_width  = cam.info
#camera_pixel  = cam.info
#println(camera_name, " ", camera_width * "x" * camera_height, " ", camera_pixel,
#  " microns")

Main.LibASICamera.set_cooler_target(cam, -9.1)
Main.LibASICamera.set_cooler_on(cam)
Main.LibASICamera.set_fan_on(cam)
Main.LibASICamera.set_exposure(cam, 1.0)
println("Cooler status: ", Main.LibASICamera.get_cooler_status(cam))
println("Exposure: ",Main.LibASICamera.get_exposure(cam))

#exit()



#
# These are additional user interface functions
#

# Set the base folder

# Set the base FITS name

# Set subframe size

# Toggle subframe on/off

# Set the image count for next image

# Set the image count maximum 

# Start a single exposure

# Start multi-frame exposures

# Stop exposure

# Set the controlled temperature

# Toggle temperature control on/off

# Input the FITS header FILTER entry

# Input the FITS header TARGET entry

# Toggle a preview display

# Exit the program


#
# User interface to the camera
#

# Set up the window

win = GtkWindow("ZWO Camera Control", 600, 200)
vbox = GtkBox(:v)
win[] = vbox

b1 = GtkButton("0")
b2 = GtkButton("2")
b3 = GtkButton("3")


b4 = GtkButton("4")
b5 = GtkButton("5")
b6 = GtkButton("6")

ab = GtkButton("Action")
mb = GtkLabel(" Message area ")
sb = GtkLabel(" Status area ")


# Example of how to modify a label

b1.label = "1"

# Create two horizontal rows of 3 buttons each

hbox1 = push!(GtkBox(:h), b1, b2, b3)
hbox2 = push!(GtkBox(:h), b4, b5, b6)

# Create a vertical sequence of rows

push!(vbox, GtkLabel("This is a label"), hbox1, hbox2, ab, mb, sb)

action_flag = false
function ab_callback(widget)
  global action_flag = !action_flag
  if action_flag
    action_flag_str = "true"
  else
    action_flag_str = "false"
  end  
  message = "Toggled action flag to " * action_flag_str
  println(message)
  mb.label = message
end


function status_callback()
  sensor_temperature = Main.LibASICamera.get_temperature(cam)
  sensor_temperature_str = string(convert(Float32, sensor_temperature))
  println(sensor_temperature_str)
  exposure = Main.LibASICamera.get_exposure(cam)
  exposure_str = string(exposure)
  status_str = "Temperature (C): " * sensor_temperature_str * " | " * "Exposure (s): " * exposure_str 
  sb.label = status_str
  return true
end

# Signal the GUI that a click on ab will run ab_callback
signal_connect(ab_callback, ab, "clicked")

# Run the status update to poll the camera
# This turns off the GUI while it is on for the given millisecond interval
# The delay must be longer that the time needed to complete update_status

g_timeout_add(status_callback, 1000)

# Start the user interface
#   in a non-interactive program run from the command line
#   uses glib to start polling the user interface

Gtk4.GLib.glib_main()

# The polling is interrupted by the timeout to do a status update
# Start an endless loop to prevent the program from closing

start_flag = true
while true

  if start_flag
    println(" Running ... ")
    global start_flag = false
  end  

  sleep(0.1)

end

exit()


#if !isinteractive()
#    c = Condition()
#    signal_connect(win, :close_request) do widget
#        notify(c)
#    end
#    @async Gtk4.GLib.glib_main()
#    wait(c)
#end
