#!/home/john/local/bin/julia

# Use an ASI ZWO camera for scientific image acquisition
#
#   Provides a GTK user interface to data acquisition
#   Exports FITS images
#

# Version 1.0: 2025-03-03 
#   Interface is actively updated with temperature and exposure data
#   Functions present to set exposure, file names, and camera temperature
#   User interface has camera status updates and displays messages
#   User interface starts exposures and reports completion
#   Libary and calls to it made consistent in use of exposure time
#   Exposure delays for completion implemented
#   Phased exposure start execution to enable display of state
#   User input of exposure time
#   Organized camera status display
#   Removed test message area
#   GUI code cleanup


###
#
# Load the ASI library wrapper with the installed ASI SDK library
#
# The file name must have an absolute directory structure
#
###

asi_image_lib_bin = "/home/john/julia/asi_image/bin/asi_image_lib.jl"
include(asi_image_lib_bin)  


# Load dependencies

using FITSIO               # FITS management
using Dates                # Date and time
using Gtk4, Gtk4.GLib      # GTK interface


# These will be altered through the user interface

exposure = 10.0
filter = " "
count = 1
count_max = 999
file_base = "image_"
file_folder = "./"
file_name = file_base * lpad(count,3,"0") * ".fits"

####
#
# Camera functions 
#
# These library functions may be used to build a camera controller
# For a complete list see the library function code
#
#   get_num_connected_cameras() returns an integer count
#   get_connected_devices() returns a list of connected devices
#   get_camera_property(id::Integer) returns ASI_CAMERA_INFO for this camera
#   get_serial_number(id::Integer) returns camera serial number
#   get_control_caps(id::Integer) returns ASI_CONTROL_CAPS for this camera 
#   get_control_value(id::Integer, control_type::ASI_CONTROL_TYPE) returns
#     a tuple (value, is_auto)
#   set_control_value(id::Integer, control_type::ASI_CONTROL_TYPE, value, 
#     auto::Bool=false)
#   set_gain(id, gain, auto=false) = 
#     set_control_value(id, ASI_GAIN, gain, auto)
#   set_exposure(id, exposure_microseconds, auto=false)
#   get_exposure(id) returns (exposure_microsecond, auto)   = get_control_value(id, ASI_EXPOSURE)  
#   set_roi_format(id::Integer, width, height, binning, img_type::ASI_IMG_TYPE)
#   get_roi_format(id::Integer)
#   set_roi_start(id::Integer, startx, starty)
#   get_roi_start(id::Integer) returns (startx, starty)
#   get_status(id::Integer)
#   open_camera(id::Integer)
#   close_camera(id::Integer)
#   init_camera(id::Integer)
#   start_exposure(id::Integer, is_dark=false)
#   stop_exposure(id::Integer)
#   get_exp_status(id::Integer) returns ASI_EXP_STATUS
#   pulse_guide_on(id::Integer, direction::ASI_GUIDE_DIRECTION)
#   pulse_guide_off(id::Integer, direction::ASI_GUIDE_DIRECTION)
#   capture_still(id::Integer) returns an array with the image
#
####

 
# Exposure status flags returned from the camera

ASI_EXP_IDLE = 0
ASI_EXP_WORKING = 1
ASI_EXP_SUCCESS = 2
ASI_EXP_FAILED = 3


# Functions to manage the camera and files

"""
    write_image_as_fits(out_file, image_data, dateobs_str, exptime_str, filter_str)

Write a FITS file containing a 16-bit image in 32-bit float with header entries

"""
function write_image_as_fits(out_file, image_data, dateobs_str, exposure_str, filter_str)
  (width, height) = size(image_data)
  f = FITS(out_file, "w")
  write(f, Float32.(image_data))
  write_key(f[1], "DATE-OBS", dateobs_str)
  write_key(f[1], "EXPTIME", exposure_str)
  write_key(f[1], "FILTER", filter_str)
  close(f)
end

"""
    capture_image(cam, exposure_seconds)

Capture an image with the ASI camera 

Accepts the exposure time in seconds 
Other camera parameters must be set before starting the exposure
Immediately starts the exposure and waits for its completion
Returns a 16-bit integer raw image


"""
function capture_image(cam, exposure)

  #width, height, binning, img_type = Main.LibASICamera.get_roi_format(cam)
  #Main.LibASICamera.set_roi_format(cam, 9576, 6388, 1, Main.LibASICamera.ASI_IMG_RAW16)
  Main.LibASICamera.set_exposure(cam, floor(Int,exposure*1000000.0), true)
  image_data = Main.LibASICamera.capture_still(cam)
  return image_data
end


"""
    snap()
    
Take a single image and saves a fits file 
Uses global parameters
Uses current camera with its settings for roi, image type, exposure 
Check camera status before calling


"""
function snap()
  exposure_str = string(exposure)
  dateobs_str = string(DateTime(now()))
  image_data = capture_image(cam, exposure)
  filter_str = filter
  write_image_as_fits(file_name, image_data, dateobs_str, exposure_str, filter_str)
  new_count = count + 1
  new_file_name = file_base * lpad(new_count,3,"0") * ".fits"
  global file_name = new_file_name
  global count = new_count
end


###
#
# Connect to a camera and set default operation parameters
#
###


devices = []
devices = Main.LibASICamera.get_connected_devices()
cam = devices[1]

# Uncomment to show the capabilities of this camera in stddout
#@show cam.info

camera_name   = cam.info.Name
camera_max_height = cam.info.MaxHeight
camera_max_width  = cam.info.MaxWidth
camera_height = camera_max_height
camera_width  = camera_max_width
camera_pixel_size  = cam.info.PixelSize

# Preset camera operation to cool and use RAW16 for lowest noise

Main.LibASICamera.set_cooler_target(cam, -9.1)
Main.LibASICamera.set_cooler_on(cam)
Main.LibASICamera.set_fan_on(cam)
Main.LibASICamera.set_exposure(cam, floor(Int,exposure*1000000.0), true)
Main.LibASICamera.set_roi_format(cam, camera_max_width, camera_max_height, 1, Main.LibASICamera.ASI_IMG_RAW16)


###
#
# The connected camera is read to use
# Build and start the GTK user interface
#
###


# User interface to the camera


# Set up the user inteface window

# Create the main window with label and size

win = GtkWindow("ZWO Camera Control", 600, 200)

# Add a vertical box to the main window

vbox = GtkBox(:v)
win[] = vbox

# Create buttons and boxes that will be added to the window

connection_box = GtkLabel(" Camera connection ")
expose_button  = GtkButton("Expose")
pause_button  = GtkButton("Pause")
multi_button  = GtkButton("Multi")
duration_button = GtkButton("Duration")
status_box = GtkLabel(" Status ")


# Create horizontal rows of buttons

hbox1 = push!(GtkBox(:h), expose_button, pause_button, multi_button)
hbox2 = push!(GtkBox(:h), duration_button)

# Add rows to the window to complete the user interface

push!(vbox, connection_box, hbox1, hbox2, status_box)

# Set up responses to the buttons

# Parse and display information on the connected camera

connection_box.label = camera_name

#  Exposure duration string input callabck
#    Sets main program exposure to the new duration
#    Sets the camera to the new duration

function set_camera_exposure_from_sec_str(new_str)
  global exposure = parse(Float64, new_str)
  exposure_microsec = floor(Int, exposure*1000000.0)
  Main.LibASICamera.set_exposure(cam, exposure_microsec, true)
end  

function duration_button_callback(widget)
  input_dialog(set_camera_exposure_from_sec_str, "Exposure time (s): ", "", win)
end

# Status update callback
#   Uses get_exp_status(id) to monitor camera 

function status_callback()
  sensor_temperature = Main.LibASICamera.get_temperature(cam)
  sensor_temperature_str = string(convert(Float32, sensor_temperature))
  exposure = Main.LibASICamera.get_exposure(cam)[1]/1000000.0
  exposure_str = string(exposure)
  status_str = "Temperature (C): " * sensor_temperature_str * "\n" * "Exposure (s): " * exposure_str 
  exp_status_str = ""
  if exposure_phase == 1
    global exposure_phase = 2
    exp_status_str = "Camera exposing"
  elseif exposure_phase == 2
    snap()
    global exposure_phase = 0          
  elseif  Main.LibASICamera.get_exp_status(cam) == ASI_EXP_IDLE
    exp_status_str = "Camera idle"
  elseif  Main.LibASICamera.get_exp_status(cam) == ASI_EXP_WORKING
    exp_status_str = "Camera exposing"
  elseif  Main.LibASICamera.get_exp_status(cam) == ASI_EXP_SUCCESS
    exp_status_str = "Exposure readout in progress"
  elseif  Main.LibASICamera.get_exp_status(cam) ==  ASI_EXP_FAILED
    exp_status_str = "Exposure failed"
  else
    exp_status_str = "Camera exposure state unknown "
  end
  status_box.label = status_str * "\n" * exp_status_str    
  return true
end

# Expose button callback
#   Uses get_exp_status(id) to monitor camera

exposure_phase = 0
function expose_button_callback(widget)
 if  Main.LibASICamera.get_exp_status(cam) == ASI_EXP_IDLE
   #snap_task = @task snap()
   #schedule(snap_task); wait(snap_task)
   #snap()
   global exposure_phase = 1
 end 
end  

# Signal the GUI that a click on expose will start an exposure
signal_connect(expose_button_callback, expose_button, "clicked")

# Signal the GUI that a click on duration will ask for the exposure duration
signal_connect(duration_button_callback, duration_button, "clicked")

# Run the status update to poll the camera
# This turns off the GUI while it is on for the given millisecond interval
# The delay must be longer that the time needed to complete update_status

g_timeout_add(status_callback, 1000)

# Start the user interface
#   in a non-interactive program run from the command line
#   uses glib to start polling the user interface

Gtk4.GLib.glib_main()

# The polling is interrupted by the timeout to do a status update
# Start an endless loop to prevent the program from closing

start_flag = true
while true

  if start_flag
    println(" Running ... ")
    global start_flag = false
  end  

  sleep(0.1)

end

exit()


#if !isinteractive()
#    c = Condition()
#    signal_connect(win, :close_request) do widget
#        notify(c)
#    end
#    @async Gtk4.GLib.glib_main()
#    wait(c)
#end
